#include "color.h"
#include "dither.h"
#include <limits>

std::unordered_map<std::string, std::vector<dither_desc> > dither_types = {
	{"none", {}},
	{"fs", {
		{1,0,7,16},
		{-1,1,3,16},
		{0,1,5,16},
		{1,1,1,16},
	}},
	{"jjn", {
		{1,0,7,48},
		{2,0,5,48},
		{-2,1,3,48},
		{-1,1,5,48},
		{0,1,7,48},
		{1,1,5,48},
		{2,1,3,48},
		{-2,2,1,48},
		{-1,2,3,48},
		{0,2,5,48},
		{1,2,3,48},
		{2,2,1,48},
	}},
	{"stucki", {
		{1,0,8,42},
		{2,0,4,42},
		{-2,1,2,42},
		{-1,1,4,42},
		{0,1,8,42},
		{1,1,4,42},
		{2,1,2,42},
		{-2,2,1,42},
		{-1,2,2,42},
		{0,2,4,42},
		{1,2,2,42},
		{2,2,1,42},
	}},
	{"atkinson", {
		{1,0,1,8},
		{2,0,1,8},
		{-1,1,1,8},
		{0,1,1,8},
		{1,1,1,8},
		{0,2,1,8},
	}},
	{"burkes", {
		{1,0,8,32},
		{2,0,4,32},
		{-2,1,2,32},
		{-1,1,4,32},
		{0,1,8,32},
		{1,1,4,32},
		{2,1,2,32},
	}},
	{"sierra", {
		{1,0,5,32},
		{2,0,3,32},
		{-2,1,2,32},
		{-1,1,4,32},
		{0,1,5,32},
		{1,1,4,32},
		{2,1,2,32},
		{-1,2,2,32},
		{0,2,3,32},
		{1,2,2,32},
	}},
	{"sierra2", {
		{1,0,4,16},
		{2,0,3,16},
		{-2,1,1,16},
		{-1,1,2,16},
		{0,1,3,16},
		{1,1,2,16},
		{2,1,1,16},
	}},
	{"sierralite", {
		{1,0,2,4},
		{-1,1,1,4},
		{0,1,1,4},
	}},
	{"fedoseev", {
		{1,0,5423,10000},
		{2,0,533,10000},
		{-2,1,246,10000},
		{-1,1,2191,10000},
		{0,1,4715,10000},
		{1,1,-23,10000},
		{2,1,-1241,10000},
		{-2,1,-65,10000},
		{-1,1,-692,10000},
		{0,1,168,10000},
		{1,1,-952,10000},
		{2,1,-304,10000},
	}},
};

void dither(png::image<png::rgb_pixel>& in, png::image<png::index_pixel>& out, const std::string colorspace, const std::string palette_mask, const std::string palette_remap, const std::string dither_type, const float error_factor) {
	auto width = in.get_width();
	auto height = in.get_height();
	auto pal = out.get_palette();
	const std::vector<dither_desc> kernel = dither_types[dither_type];
	png::pixel_buffer<hdr_color> err(width,height);

	auto pmask = intset(palette_mask);
	auto remap = intmap(palette_remap);

	hdr_color oldpixel, newpixel;
	for(int y = 0; y<height; y++) {
		for(int x = 0; x<width; x++) {
			oldpixel = clamp(add(from(in[y][x]),err[y][x]));
			size_t c_i = closest(pal, oldpixel, colorspace, pmask);
			newpixel = from(pal[c_i]);
			if(remap.count(c_i) > 0) {
				out[y][x] = remap[c_i];
			} else {
				out[y][x] = c_i;
			}

			hdr_color epsilon = diff(oldpixel, newpixel);
			for(auto k : kernel) {
				int kx = modulo(x+k.x,width);
				int ky = modulo(y+k.y,height);
				err[ky][kx] = add(err[ky][kx], multi(multidiv(epsilon, k.num, k.den), error_factor));
			}
		}
	}
}
